from scipy.stats import ttest_ind
from sklearn.ensemble import RandomForestClassifier
from sklearn.linear_model import LogisticRegression
from utils import *
from evaluation import *
import numpy as np
import matplotlib.pyplot as plt
import itertools

# import rpy2.robjects as ro
# from rpy2.robjects.numpy2ri import numpy2ri

# # Some functions for scoring how good a split is
# def sigclust(X,y):
#     if len(np.unique(y)) != 2: 
#         print 'ERROR: can only have 2 unique labels'
#         return
#     y = str_labels_to_ints(y)
#     ro.r('library(sigclust)')
#     ro.r('''g<-function(X,y){
#                 p = sigclust(X,100,labflag=1,label=y,icovest=3)
#                 return(p@pval)
#             }''')
#     g = ro.globalenv['g']
#     p = g(numpy2ri(X),numpy2ri(y))
#     return np.nan_to_num(-np.log10(p[0]))

# Feature selection using random forest 
def skRandomForest(X,Y):
    clf = RandomForestClassifier()
    clf.fit(X,Y)
    return clf

def select_genes_using_RF(X,Y,return_score=False,verbose=False,return_comparisons=False,score_function=None):
    # find most important features using random forest
    n = 100 # number of times to repeat random forest
    m = int(np.round(X.shape[1]/2)) # number of features to output at the end
    if m > n: m = n
    # Save the number of times each gene appears in the top n most important features
    gene_scores = np.zeros(X.shape[1])
    # For each fitting, score how well the fitting classifies the data
    if return_score and score_function is None: scores = np.zeros(n)
    for i in range(0,n):    
        clf = skRandomForest(X,Y)
        importances = clf.feature_importances_
        y_clf = clf.predict(X)
        if return_score and score_function is None: scores[i] = NMI(Y,y_clf)
        gene_ranks = np.flipud(np.argsort(importances))
        gene_scores[gene_ranks[0:m]] += 1
    # Rank genes by how often it appeared in the top n most important features
    gene_ranks = np.flipud(np.transpose(np.argsort(np.transpose(gene_scores))))[0:m]
    gene_scores /= float(n)
    outputs = [gene_ranks,gene_scores[gene_ranks]]
    if verbose: print 'Mean prediction score: %.3f'%(np.mean(scores))
    if return_score: 
        if score_function is None: outputs.append(np.mean(scores))
        else: outputs.append(score_function(X,Y))
    if return_comparisons: outputs.append(compare_feature_means(X,Y)[gene_ranks])
    return tuple(outputs)

def log_select_genes_using_RF(X,Y,**kwargs):
    return select_genes_using_RF(np.log(1+X),Y,**kwargs)

def log_select_genes_using_RF_sigclust(X,Y,**kwargs):
    return select_genes_using_RF(np.log(1+X),Y,score_function=sigclust,**kwargs)

# Feature selection using L1-regularized logistic regression
def skLogisticRegression(X,Y):
    lr = LogisticRegression(penalty='l1')
    lr.fit(X,Y)
    return lr

def select_genes_using_LR(X,Y,return_score=False,verbose=False,return_comparisons=False,score_function=None):
    clf = skLogisticRegression(X,Y)
    y_clf = clf.predict(X)
    gene_scores = np.abs(clf.coef_)[0]
    gene_ranks = np.flipud(np.argsort(gene_scores))
    outputs = [gene_ranks,gene_scores[gene_ranks]]
    if verbose: print 'Score: %.2f'%(score)
    if return_score: 
        if score_function is None: score = compute_clustering_accuracy(Y,y_clf)
        else: score = score_function(X,Y)
        outputs.append(score)
    if return_comparisons: outputs.append(compare_feature_means(X,Y)[gene_ranks])
    return tuple(outputs)

def log_select_genes_using_LR(X,Y,**kwargs):
    return select_genes_using_LR(np.log(1+X),Y,**kwargs)

def log_select_genes_using_LR_sigclust(X,Y,**kwargs):
    return select_genes_using_LR(np.log(1+X),Y,score_function=sigclust,**kwargs)

# Feature selection using Welch's t-test
def select_genes_using_Welchs(X,Y,return_score=False,verbose=False,return_comparisons=False,
                              score_function=None,return_tp=False,equal_var=False):
    if len(np.unique(Y)) > 2: 
        print 'ERROR: Y should only have 2 unique values'
    N,M = np.shape(X)
    y = str_labels_to_ints(Y)
    X1,X0 = X[y==1,:],X[y==0,:]
    # Numerical error edge case handling
    epsilon = 1e-8
    num_err_fix = np.abs(np.mean(X1,0)-np.mean(X0,0)) < epsilon
    X1[:,num_err_fix] = 0
    X0[:,num_err_fix] = 0
    # Perform t-test
    t_orig,p_orig = ttest_ind(X1,X0,equal_var=equal_var)
    gene_inds = np.array(range(M))
    finite_inds = np.isfinite(t_orig) # To account for cases where gene has 0 expression in both populations
    t,p,gene_inds = t_orig[finite_inds],p_orig[finite_inds],gene_inds[finite_inds]
    # Argsort first by p-value, then by -np.abs(t-value)
    keep_inds = np.array([j[2] for j in sorted([(p[i],-np.abs(t[i]),i) for i in range(len(t))])])
    gene_ranks = gene_inds[keep_inds]
    gene_scores = np.nan_to_num(-np.log10(p[keep_inds]))
    outputs = [gene_ranks,gene_scores]
    if verbose: print 'Score: %.2f'%(score)
    if return_score: 
        if score_function is None: score = gene_scores[0]
        else: score = score_function(X,Y)
        outputs.append(score)
    if return_comparisons: outputs.append((t>0).astype(int)[keep_inds])
    if return_tp: 
        outputs.append(t_orig)
        outputs.append(p_orig)
    # Note: For dendrosplit, each split involves exactly two clusters. 'L' will be mapped
    #   to 0, and 'R' will mapped to 1. A positive t statistic indicates that the 1-clust
    #   has a higher median than the 0-clust (i.e. 'L' has less of the feature than 'R' on
    #   average). Therefore return_comparisons will index the features that are more greater
    #   in 1-clust (or 'R'-clust). In the visualize_history function, 0's are red, 1's are
    #   green.
    return tuple(outputs)

def log_select_genes_using_Welchs(X,Y,**kwargs):
    return select_genes_using_Welchs(np.log(1+X),Y,**kwargs)

def log_select_genes_using_Welchs_sigclust(X,Y,**kwargs):
    return select_genes_using_Welchs(np.log(1+X),Y,score_function=sigclust,**kwargs)

# Visualization scripts
def one_from_rest_gene_selection(X,Y,feature_selector=log_select_genes_using_Welchs):
    selected_genes_for_each_cluster = []
    for ctype in np.unique(Y):
        selected_genes_for_each_cluster.append(feature_selector(X,Y == ctype,return_score=True))
    return selected_genes_for_each_cluster
        
def one_from_rest_visualize_genes(X,genes,x1,x2,y,num_genes=3,
                                  feature_selector=log_select_genes_using_Welchs):
    selected_genes_for_each_cluster = one_from_rest_gene_selection(X,y,feature_selector)
    unique_clusts = np.unique(y)
    for k,selected_genes in enumerate(selected_genes_for_each_cluster):
        plt.figure(figsize=(4*(num_genes+1),3))
        plt.subplot(1,num_genes+1,1)
        plt.scatter(x1,x2,c=y==unique_clusts[k],edgecolors='none')
        plt.title('Clust: %d, Acc: '%(unique_clusts[k])+sn(selected_genes[2]))
        _ = plt.axis('off')
        for i in range(num_genes):
            g = genes[selected_genes[0][i]]
            plt.subplot(1,num_genes+1,i+2)
            plt.scatter(x1,x2,c=X[:,genes == g],edgecolors='none')
            plt.title(g+' '+sn(selected_genes[1][i]))      
            _ = plt.axis('off')

# Note that if you use save_name while show_plots is True, all the plots will be saved as well
def save_more_highly_expressed_genes_in_one_clust(X,genes,y,x1=None,x2=None,num_genes=3,verbose=True,
                                                  feature_selector=log_select_genes_using_Welchs,save_name=None,
                                                  show_plots=True,pval_cutoff=10):
    if show_plots:
        if x1 is None or x2 is None: print 'NEED TO PASS IN x1, x2 FOR PLOTTING'
    if save_name is not None: 
        f = open(save_name+'_cluster_features.txt','w')
        print >> f, 'cluster\tgene\tpvalue\tmean_of_expr\tfold_change_of_expr'
    for c in np.unique(y):
        c1 = X[y == c,:]
        c2 = X[y != c,:]
        keep_genes = np.mean(c1,0) > np.mean(c2,0) 
        g_temp = genes[keep_genes]
        if len(g_temp) == 0: continue
        if verbose: print 'Cluster %d: %d/%d genes kept'%(c,np.sum(keep_genes),len(keep_genes))
        num_genes_c = np.min([np.sum(keep_genes),num_genes])
        gene_ranks,gene_scores,score = feature_selector(X[:,keep_genes],y==c,return_score=True)
        if show_plots and x1 is not None and x2 is not None:
            plt.figure(figsize=(4*(num_genes_c+1),3))
            plt.subplot(1,num_genes_c+1,1)
            plt.scatter(x1,x2,c=y==c,edgecolors='none')
            plt.title('Clust: %d, Score: '%(c)+sn(score))
            _ = plt.axis('off')
        for i in range(num_genes_c):
            g = g_temp[gene_ranks[i]]
            g_ind = np.where(genes==g_temp[gene_ranks[i]])[0]
            if show_plots and x1 is not None and x2 is not None:
                plt.subplot(1,num_genes_c+1,i+2)
                plt.scatter(x1,x2,c=X[:,genes==g],edgecolors='none')
                plt.title(g+' '+sn(gene_scores[i]))
                _ = plt.axis('off')
            if save_name is not None and gene_scores[i] > pval_cutoff:
                g_mean_in = np.mean(X[y==c,g_ind])
                g_mean_out = np.mean(X[y!=c,g_ind])
                s = '\t'.join([str(c),g_temp[gene_ranks[i]],sn(gene_scores[i],j=10),
                               sn(g_mean_in,j=10),sn(g_mean_in/g_mean_out,j=10)])
                print >> f, s
        if show_plots and x1 is not None and x2 is not None and save_name is not None: 
            plt.savefig(save_name+'_cluster'+str(c)+'.png', format='png', dpi=300)
    if save_name is not None: f.close()

# Note that if you use save_name while show_plots is True, all the plots will be saved as well
def pairwise_cluster_comparison(X,genes,y,x1=None,x2=None,num_genes=20,verbose=True,
                                feature_selector=log_select_genes_using_Welchs,save_name=None,
                                show_plots=True,pval_cutoff=10):
    if show_plots:
        if x1 is None or x2 is None: print 'NEED TO PASS IN x1, x2 FOR PLOTTING'
    if save_name is not None:
        f = open(save_name+'_pairwise_cluster_features.txt','w')
        print >> f, 'c1\tc2\tgene\tpvalue\tlarger_clust\tfold_change_of_exp_for_larger_clust'
    for (i,j) in itertools.combinations(np.unique(y),2):
        labels_ij = np.logical_or(y==i,y==j)
        gene_ranks,gene_scores,score = feature_selector(X[labels_ij,:],y[labels_ij],return_score=True)
        if verbose: print '\nComparing cluster %s and %s'%(i,j)
        if show_plots and x1 is not None and x2 is not None:
            plt.figure(figsize=(4*(num_genes+1),3))
            plt.subplot(1,num_genes+1,1)
            plot_labels_legend(x1[labels_ij],x2[labels_ij],y[labels_ij])
            plt.title('Clust %s v. clust %s, Score: '%(i,j)+sn(score))
            _ = plt.axis('off')
        for k in range(np.min([len(gene_ranks),num_genes])):
            g_mean_i = np.mean(X[y==i,gene_ranks[k]])
            g_mean_j = np.mean(X[y==j,gene_ranks[k]])
            if g_mean_i > g_mean_j: 
                c = i
                fold = g_mean_i/g_mean_j
            else: 
                c = j
                fold = g_mean_j/g_mean_i
            if verbose: print '%s (more expressed in %s) '%(genes[gene_ranks[k]],c)+sn(gene_scores[k])
            if show_plots and x1 is not None and x2 is not None:
                plt.subplot(1,num_genes+1,k+2)
                plt.scatter(x1[labels_ij],x2[labels_ij],c=X[labels_ij,gene_ranks[k]],edgecolors='none')
                plt.title(genes[gene_ranks[k]]+' '+sn(gene_scores[k]))
                _ = plt.axis('off')
                if save_name is not None: plt.savefig()
            if save_name is not None and gene_scores[k] > pval_cutoff:
                s = '\t'.join([str(i),str(j),genes[gene_ranks[k]],
                               sn(gene_scores[k],j=10),str(c),sn(fold,j=10)])
                print >> f, s
        if show_plots and x1 is not None and x2 is not None and save_name is not None:
            plt.savefig(save_name+'_c'+str(i)+'_v_c'+str(j)+'.png', format='png', dpi=300)  



import random,string,pyclust,itertools,sys,os
import numpy as np
from sklearn.metrics.pairwise import pairwise_distances
from sklearn.cluster import SpectralClustering,DBSCAN,KMeans,AgglomerativeClustering,AffinityPropagation
from sklearn.neighbors import kneighbors_graph

# CLUSTERING ALGORITHMS

# eps: maximum distance allowed between two points in the same neighborhood
# min_samples: # of samples in a neighborhood for a point to be considered "core"
def skDBSCAN(D,eps=0.05,min_samples=5):
    dbscan = DBSCAN(eps=eps,min_samples=min_samples,metric='precomputed')
    return dbscan.fit_predict(D)

# D NEEDS TO BE A SIMILARITY/AFFINITY MATRIX 
# pref: as preference increases, so does the # of exemplars (and # of clusters)
# damp: higher damping = higher weight on older messages 
def AffinityProp(D,pref,damp):
    aff=AffinityPropagation(affinity='precomputed',preference=pref,damping=damp)
    labels=aff.fit_predict(D)
    return labels

# hierarchical clustering with connectivity information
def hclust(D,k,compute_connectivity_info = False):
    if compute_connectivity_info:
        knn_graph = kneighbors_graph(D, 7, include_self=False, metric='precomputed')
        hc = AgglomerativeClustering(n_clusters=k,connectivity=knn_graph,
                                     affinity='precomputed',linkage='complete')
    else:
        hc = AgglomerativeClustering(n_clusters=k,affinity='precomputed',linkage='complete')
    return hc.fit_predict(D)

# This one is a bit different. Cannot use a distance matrix
def kMedoids(X,k,d_metric='euclidean'):
    km = pyclust.KMedoids(n_clusters=k,distance=d_metric)
    return km.fit_predict(X)

# Kmeans
def kMeans(X,k):
    km = KMeans(n_clusters=k)
    return km.fit_predict(X)

# SNN-Cliq clustering algorithm
def SNN_Cliq(D,r=0.7,m=0.5,k=5,save_dist=True,verbose=False,deletelabels=True,SNNdir=None):
    if SNNdir is None:
        print 'Need to set SNNdir to the directory where SNN.R and Cliq.py are saved'
        print '(see directory where this module is saved)'
        return
    suffix = generate_random_string(10)
    distfile='tempDistFile_'+suffix
    edgefile='tempEdgeFile_'+suffix
    labelsfile='labels_'+str(k)+'_'+suffix
    if save_dist: np.savetxt(SNNdir+distfile,D)
    start_time = time.time()
    if verbose: print 'Building SNN graph..'
    os.system('Rscript '+SNNdir+'SNN.R '+SNNdir+distfile+' '+SNNdir+edgefile+' '+str(k))
    if verbose: print 'SNN graph built (%.3f s)'%(time.time()-start_time)
    start_time = time.time()
    if verbose: print 'Finding clusters..'
    os.system('python '+SNNdir+'Cliq.py -i '+SNNdir+edgefile+' -o '+SNNdir+labelsfile
              +' -r '+str(r)+' -m '+str(m)+' -n '+str(len(D)))
    if verbose: print 'Clusters found (%.3f s)'%(time.time()-start_time)
    os.system('rm '+SNNdir+distfile+' '+SNNdir+edgefile)
    y = np.loadtxt(SNNdir+labelsfile,dtype=int)
    os.system('rm '+SNNdir+labelsfile)
    return y

def generate_random_string(N):
    return ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(N))

# ------------------------------------------------------------------------------
# CONSENSUS

# all_labels list of labels. This function finds a consensus amongst them by counting how
# often two points are put in the same cluster
def build_consensus_matrix(all_labels):
    n = len(all_labels[0])
    D = np.zeros((n,n))
    cIDs = np.array(range(n))
    for i,labels in enumerate(all_labels):
        sys.stdout.write("\rbuilding consensus matrix.. %d/%d"%(i+1,len(all_labels)))
        sys.stdout.flush()
        D = dist_update(D,labels,cIDs,np.max(labels)+1)
    # Map similarity matrix to distance matrix
    D = D/float(len(all_labels)) + np.diag(np.ones(len(D))) 
    return 1-D

# function to update similarity matrix counts based on which samples are in the same cluster
def dist_update(D,y,cIDs,k):
    for clustID in range(k):
        for i,j in itertools.combinations(cIDs[y == clustID],2):
            D[i,j] += 1
            D[j,i] += 1
    return D



import colorsys
import numpy as np
import matplotlib.pyplot as plt

# Load csv file (skipcol = 4 for resolve)
def load_csv(filename,skipcol=4):
    X = np.genfromtxt(filename,delimiter=',',skip_header=1)
    X = X[:,skipcol:]
    print 'We have %d samples and %d features per sample.'%np.shape(X)
    print '%.3f%% of the entries are 0.'%(100*np.sum(X == 0)/float(np.prod(np.shape(X))))
    return X

# Check if all entries in a numpy matrix are whole numbers
def all_entries_are_whole_nums(X):
    a = np.product(np.shape(X))
    b = sum([1 for i in np.ndarray.flatten(X) if float.is_integer(i)])
    return a == b

# When printing, decide whether to use scientific notation (if value gets too big)
def sn(i,j=2):
    if i > 1000 or i < 0.01: return '%.1E'%(i)
    s = '%.'+str(j)+'f'
    return s%(i)

# Simple way to plot colors and labels with legend                                        
#  -if x1, x2 is just a subsample of all the points, need select_inds to indicate
#   which points they correspond to
def plot_labels_legend(x1,x2,y_,labels=None,save_name=None,label_singletons=True,
                       show_axes=False,xlog_scale=False,ylog_scale=False,
                       legend_pos=(1.4, 1.0),markersize=5,select_inds=None):

    if select_inds is not None: y = y_[select_inds]
    else: y = y_ 

    N = len(np.unique(y))
    Ncolors = N+1
    if label_singletons:
        for i,c in enumerate(np.unique(y)):
            if np.sum(y == c) == 1: 
                y[y == c] = -1
                Ncolors -= 1
    if labels == None: labels = np.unique(y)
    if -1 in labels or '-1' in labels: Ncolors -= 1
    HSVs = [(x*1.0/Ncolors, 0.8, 0.9) for x in range(Ncolors)]
    RGBs = map(lambda x: colorsys.hsv_to_rgb(*x), HSVs)
    j = 0
    for i in labels:
        if i != -1 and i != '-1':
            plt.plot(x1[y==i],x2[y==i],'.',c=RGBs[j],label=str(i)+' ('+str(np.sum(y_==i))+')',
                     markersize=markersize)
            j += 1
        else:
            plt.plot(x1[y==i],x2[y==i],'*',c='w',label='Singletons ('+str(np.sum(y_==i))+')',
                     markersize=markersize,markeredgecolor='k')
    if not show_axes: _ = plt.axis('off')
    if xlog_scale: plt.xscale('log')
    if ylog_scale: plt.yscale('log')
    if legend_pos is not None: plt.legend(bbox_to_anchor=legend_pos)
    if save_name is not None: plt.savefig(save_name+'.png', format='png', dpi=300)

# For each feature, determine the index of the cluster with the greatest expression
def compare_feature_means(X,Y):
    Nc = len(np.unique(Y))
    N,M = np.shape(X)
    m = np.zeros((Nc,M))
    for i,c in enumerate(np.unique(Y)):
        m[i,:] = np.mean(X[Y == c,:],0)
    return np.argmax(m,0)

# Map labels to integers
def str_labels_to_ints(y_str):
    y_int = np.zeros(len(y_str))
    for i,label in enumerate(np.unique(y_str)):
        y_int[y_str == label] = i
    return y_int.astype(int)

# Get all off-diagonal entries in a distance matrix
def flatten_distance_matrix(D,inds=None):
    if inds is not None: D2 = cut_matrix_along_both_axes(D,inds)
    else: D2 = D
    d = D2.reshape(-1,1)
    return np.delete(d,np.where(d==0)[0])

# index a 2D matrix along both axes
def cut_matrix_along_both_axes(X,inds):
    Z = X[inds,:]
    return Z[:,inds]




from scipy.cluster.hierarchy import dendrogram, linkage
from scipy.spatial.distance import squareform
import numpy as np
import matplotlib.pyplot as plt
import time,colorsys

from preprocessing import *
from feature_selection import *
from utils import *

# Using the linkage matrix Z, iteratively perform the following from the top:
#    1. Split into two clusters
#    2. If one of the clusters are too small (< min_clust_size), mark all points in it as outliers/error. remove it
#    3. Else, use random forest to classify the two types against each other. If high enough accuracy, keep clusters
#    4. Repeat on subclusters that aren't marked as outliers/error

# Given an inner node, split the two subtrees into two groups 
def node_to_leaf_labels(Z,i,ltypes=['0','1'],defaultlabel='_'):
    # Z: hierarchical clustering result returned by scipy linkage function
    # i: row of Z to perform splitting on 
    # ltypes: 2x1 vector. labels given to subtrees
    n = int(Z[-1,3])
    labels = [defaultlabel for _ in range(n)]
    def recursion(j,label):
        # base case 1: leaf node
        if j < n: 
            labels[j] = label
        # base case 2: parent of two leaf nodes
        elif Z[j-n,3] == 2:
            labels[int(Z[j-n,0])] = label
            labels[int(Z[j-n,1])] = label
        # recursive case: 
        else:
            recursion(int(Z[j-n,0]),label)
            recursion(int(Z[j-n,1]),label)
    recursion(int(Z[i,0]),ltypes[0])
    recursion(int(Z[i,1]),ltypes[1])
    return np.array(labels)

# Label updating algorithm for dendrosplit
def update_labels(y,labels,background='_'):
    label = None
    for i in range(len(y)):
        if y[i] != background:
            if label == None: label = labels[i]
            labels[i] += y[i] # Update labels 
    return label,labels

# Dendrosplit algorithm
def dendrosplit(X,preprocessing=log_correlation,min_clust_size=2,score_threshold=10,
                method='complete',verbose=False,split_evaluator=log_select_genes_using_Welchs,
                disband_percentile=100):
    '''
        Perform iterative hierarchical clustering. Obtain a dendrogram using a distance matrix. 
        Starting from the top, iteratively split the tree into two subtrees. Perform a test to 
        see if the two subtrees are different enough.
        
        Inputs:
            X: Matrix of counts 
            preprocessing: function to convert X to a distance matrix. This argument
                can be set to 'precomputed', in which case X should be a tuple (D,X)
                where D is the already-computed distance matrix
            min_clust_size: min size of cluster in order for a cluster split to be considered
            score_threshold: min score achieved by split required to keep the split
            method: linkage method for hierarchical clustering
            verbose: whether or not to turn on print statements while the algorithm runs
            split_evaluator: function to score how good a split is
            disband_percentile: if both subtrees have less than min_clust_size samples or if 
                a candidate split does not achieve high enough a score, look at the pairwise 
                distances amongst samples in this final cluster. if all are greater than 
                this percentile of the distances, then mark all of the points as singletons
            
        Outputs:
            y: cluster labels 
            history: list of tuples of information generated at each split. Use the functions
                'print_history' and 'visualize_history' with this output
    '''
    
    start_time = time.time()
    
    # Catch edge cases
    if preprocessing != 'precomputed': 
        if np.sum(np.sum(X,0) == 0) > 1:
            print 'Please remove genes that sum to 0 (try split.filter_genes())'
            return
#         if not check_if_all_entries_are_whole(X):
#             print 'Please make sure X is a valid distance matrix'
#             return
        D = preprocessing(X)
        if verbose: print 'Preprocessing took %.3f s'%(time.time()-start_time)
    elif type(X) != tuple: 
        print 'preprocessing = precomputed. Feed in (D,X) tuple as first argument.'
        return
    elif not np.all(np.isclose(X[0],X[0].T)):
        print 'Need a valid distance matrix.'
        return
    else: 
        D,X = X[0],X[1]
        if np.sum(np.sum(X,0) == 0) > 1:
            print 'Please remove columns that sum to 0 (try split.filter_genes())'
            return
#         if not check_if_all_entries_are_whole(X):
#             print 'Please make sure X is a valid distance matrix'
#             return

    # Perform the hierarchical clustering
#    D = (D+D.T)/2
    Ds = squareform(D)
    Z = linkage(Ds,method=method)

    N = len(X)

    # Compute disband percentile
    disband_threshold = np.percentile(flatten_distance_matrix(D),disband_percentile)
    
    # start with the label 'r' (for 'root') for all samples
    labels = ['r' for i in range(N)]

    # Stuff we may want to remember about each stage of splitting:
    #  - labels before the split and # of labels
    #  - labels after the split and # of each labels
    #  - splitting score
    #  - important features for distingushing the two subtypes
    #  - which cluster has greater expression of each feature
    history,scount = [],[0]
    
    def recursion(j,split_depth,labels):

        j = int(j)

        # Check number of leaves below node exceeds required threshold
        # Or if we're at a leaf node
        if j < 0: return
        if Z[j,3] < min_clust_size: return
        
        # Split subtrees and get labels
        y_node = node_to_leaf_labels(Z,j,ltypes=['L','R'])
      
        lsize = np.sum(y_node == 'L')
        rsize = np.sum(y_node == 'R')
        if verbose: print 'Potential split result: %d and %d'%(lsize,rsize)
        
        # If the split produces a cluster that's too small, ignore the smaller cluster
        if lsize < min_clust_size and rsize < min_clust_size: 
            # if both clusters are too small, check distances. If all distances are above
            # the disband_percentile overall distance, then make all samples singletons
            if np.sum(flatten_distance_matrix(D,y_node!='_') < disband_threshold) == 0:
                if verbose: print 'Disbanding (points in cluster too far from each other)'
                for i in range(N):
                    # guarantee that each sample has a unique label
                    if y_node[i] != '_': labels[i] += y_node[i]+str(i)
            return

        if lsize < min_clust_size: 
            label,labels = update_labels(y_node,labels)
            history.append(({label: np.sum(y_node != '_'),
                             label+'L': np.sum(y_node == 'L'),
                             label+'R': np.sum(y_node == 'R')},
                            y_node,Z[j,2],split_depth))
            recursion(Z[j,1]-N,split_depth,labels)
            return
        if rsize < min_clust_size: 
            label,labels = update_labels(y_node,labels)
            history.append(({label: np.sum(y_node != '_'),
                             label+'L':np.sum(y_node == 'L'),
                             label+'R':np.sum(y_node == 'R')},
                            y_node,Z[j,2],split_depth))
            recursion(Z[j,0]-N,split_depth,labels)
            return
        
        # Given a data matrix and some labels, fit a classifier. Evaluate classification accuracy
        scount[0] += 1
        genes,rankings,score,comparisons = split_evaluator(X[y_node!='_',:],y_node[y_node!='_'],
                                                           return_score=True,return_comparisons=True)
        if verbose: print ' Split score '+sn(score)

        # Decide if we should keep the split based on the score
        if score > score_threshold:
            label,labels = update_labels(y_node,labels)
            history.append(({label: np.sum(y_node != '_'),
                             label+'L': np.sum(y_node == 'L'),
                             label+'R': np.sum(y_node == 'R')},
                             y_node, Z[j,2], split_depth+1,
                             genes, rankings, score, comparisons))
            recursion(Z[j,0]-N,split_depth+1,labels)
            recursion(Z[j,1]-N,split_depth+1,labels)
        else: 
            if np.sum(flatten_distance_matrix(D,y_node!='_') < disband_threshold) == 0:
                if verbose: print 'Disbanding (points in cluster too far from each other)'
                for i in range(N):
                    if y_node[i] != '_': labels[i] += y_node[i]+str(i)
            return
        
    recursion(len(Z)-1,0,labels)
    if verbose: 
        print '# of times score function was called: %d'%(scount[0])
        print 'Total computational time was %.3f s'%(time.time()-start_time)
    return np.array(labels),history

# calculate full dendrogram
def plot_dendro(D,sample_names=None,p=None,labels=None,save_name=None,method='complete',lr=90,return_cell_order=False,fig_dim=(100,10),font_size=7):
    '''
    Calculate the full dendrogram based on a distance matrix D
    
    Inputs: 
    D: distance matrix (NxN)
    sample_names: name for each sample (uses the index by default, length-N vector)
    p: depth to cut the tree (goes to leaves by default)
    labels: a label for each sample (also a length-N vector, good for indicating cluster IDs)
    method: linkage methods for scipy's hierarchical clustering algorithm
    lr: leaf rotation
    return_cell_order: boolean. If true, returns the indices of the samples in the order of the dendrogram
    fig_dim: tuple indicating the size of the figure
    font_size: font size of leaves (7 by default)
    '''
    if p is None: p = len(D)
    if sample_names is None: sample_names = range(len(D))
    if labels is not None:
        Ncolors = len(np.unique(labels))
        HSVs = [(x*1.0/Ncolors, 0.8, 0.9) for x in range(Ncolors)]
        RGBs = map(lambda x: colorsys.hsv_to_rgb(*x), HSVs)
        color_map = {i:RGBs[j] for j,i in enumerate(np.unique(labels))}
        colors = map(lambda x:color_map[x],labels)
    # Convert n-by-n D to (n choose 2)-by-1 vector for scipy's linkage function
    Ds = squareform(D)
    # 'complete' merges clusters based on farthest points in two clusters
    # 'single' merges clusters based on closest points in two clusters
    # 'average' merges clusters based on the average pairwise distance between all points in two clusters
    # 'weighted' merges clusters A and B based on the average distance of the two clusters that were merged to become A
    Z = linkage(Ds,method=method)
    plt.figure(figsize=fig_dim)
    D = dendrogram(Z,truncate_mode='lastp',p=p,leaf_rotation=lr,leaf_font_size=font_size,labels=sample_names,)
    ax = plt.gca()
    if labels != None:
        xlbls = ax.get_xmajorticklabels()
        for i in range(len(xlbls)):
            xlbls[i].set_color(colors[int(float(xlbls[i].get_text()[0:]))])
    if save_name is not None: plt.savefig(save_name+'.png', format='png', dpi=300)
    if return_cell_order: return np.array([int(i.get_text()) for i in ax.get_xmajorticklabels()])

        
# Visualize how the splitting progresses
def visualize_history(X,x1,x2,genes,history,save_name=None,axisoff=False,
                      markersize=5,select_inds=None,thresh=0,save_separately=False):
    prefix = 'Split'
    ii = 1
    if type(history[0]) != tuple: 
        history = history[2:]
        prefix = 'Merge'
    for i in history: 
        if len(i) < 8: continue
        if i[-2] < thresh: continue
        if save_separately:
            plt.figure(figsize=(5,5))
        else:
            plt.figure(figsize=(16,4))
            plt.subplot(1,3,1)
        plot_labels_legend(x1,x2,str_labels_to_ints(i[1]),markersize=markersize,select_inds=select_inds)
        plt.title('%s %d: '%(prefix,ii)+'  '.join(['%s(%d): '%(genes[i[4][j]],i[7][j])+sn(i[5][j]) for j in range(len(i[7])) if j < 3]))
        if save_separately:
            if save_name is not None:
                plt.savefig(save_name+'_candidateclusts_'+str(ii)+'.png', format='png', dpi=300, bbox_inches='tight')
            plt.figure(figsize=(5,5))
        else: plt.subplot(1,3,2)
        if select_inds is not None:
            expr = X[select_inds,i[4][0]]
        else: expr = X[:,i[4][0]]
        plt.scatter(x1,x2,edgecolors='none',c=expr)
        plt.title('Expression of '+genes[i[4][0]])
        _ = plt.axis('off')
        if save_separately:
            if save_name is not None:
                plt.savefig(save_name+'_expr_'+str(ii)+'.png', format='png', dpi=300, bbox_inches='tight')
            plt.figure(figsize=(5,5))
        else:
            plt.subplot(1,3,3)
        v0 = X[i[1]=='L',i[4][0]]
        v1 = X[i[1]=='R',i[4][0]]
        bins = np.histogram(np.hstack((v0,v1)), bins=20)[1]
        data = [v0,v1]
        plt.hist(data,bins,label=['0','1'],alpha=0.8,color=['r','g'],normed=True,edgecolor='none')
        plt.xlabel('expression')
        if axisoff: plt.axis('off')
        plt.title('Expression of '+genes[i[4][0]])
        ii += 1
        if save_name is not None: 
            if save_separately:
                plt.savefig(save_name+'_exprhist_'+str(ii)+'.png', format='png', dpi=300, bbox_inches='tight')
            else:
                plt.savefig(save_name+'_'+str(ii)+'.png', format='png', dpi=300, bbox_inches='tight')
        
# Print the history in an interpretable way
def print_history(genes,history):
    prefix = 'Pre-split'
    if type(history[0]) != tuple:
        print '%d of %d samples are singletons'%(np.sum(np.bincount(str_labels_to_ints(history[0]))==1),len(history[0]))
        print_singleton_merging_result(history[0],history[1])
        history = history[2:]
        prefix = 'Post-merge'
    for i in history:
        if len(i) < 8: continue
        if type(i[0]) == dict:
            nP = max(i[0].values())
            for key in i[0]:
                if i[0][key] < nP and key[-1] == 'L': nL = i[0][key]
                if i[0][key] < nP and key[-1] == 'R': nR = i[0][key]
            print prefix+": %-5s  L: %-5s  R: %-5s  Score: "%(nP,nL,nR)+ \
                  sn(i[6])+"  Top Gene: %-10.10s  Top Gene Score: "%(genes[i[4][0]])+sn(i[5][0])

# Print more information on how singletons were handled
def print_singleton_merging_result(L1,L2):
    # Make sure L1 is the result of merging singletons (has less unique clusters)
    if len(np.unique(L1)) > len(np.unique(L2)): L1,L2 = L2,L1
    outliers = [str(i) for i in np.where(L1 == -1)[0]]
    print 'Singleton(s) '+', '.join(outliers)+' marked as outliers (N = %d)'%(len(outliers))
    for c in np.unique(L1):
        if c != -1:
            s = ''
            large_clust = None
            Ltemp = L2[L1 == c]
            if len(np.unique(Ltemp)) > 1:
                for j in np.unique(Ltemp): 
                    if np.sum(Ltemp == j) == 1: s += str(j)+', ' 
                    else: large_clust = j
                if large_clust is not None:
                    print 'Singleton(s) '+s[:-2]+' merged with cluster %d (N = %d) to form cluster %d (N = %d)' \
                          %(large_clust,np.sum(Ltemp == large_clust),c,np.sum(L1==c))
                else:
                    print 'Singleton(s) '+s[:-2]+' merged to form cluster %d (N = %d)' \
                          %(c,np.sum(L1==c))

# Analyze a split
def analyze_split(X,x1,x2,genes,history,split_num,num_genes=12,clust=None,show_background=True):
    prefix = 'Split'
    if type(history[0]) != tuple: 
        history = history[2:]
        prefix = 'Merge'

    ii = 0
    for k,i in enumerate(history):
        if len(i) == 8: ii += 1
        if ii == split_num: break

    i = history[k]
    y = str_labels_to_ints(i[1])
    important_gene_indices = i[4]
    scores = i[5]

    comparisons= i[7]

    if clust is not None:
        if num_genes > np.sum(comparisons == clust): 
            print 'num_genes is greater than the number of genes more highly expressed in clust %d'%(clust)
            num_genes = np.sum(comparisons == clust)
        important_gene_indices = important_gene_indices[comparisons == clust]
        scores = scores[comparisons == clust]
        comparisons = comparisons[comparisons == clust]

    plot_labels_legend(x1,x2,y)

    if show_background is False:
        x1 = x1[y != 2]
        x2 = x2[y != 2]
        X = X[y != 2,:]
        
    plt.title('%s %d: '%(prefix,split_num))
    for j in range(num_genes):
        if j%4 == 0: plt.figure(figsize=(16,4))
        plt.subplot(1,4,j%4+1)
        plt.scatter(x1,x2,edgecolors='none',c=X[:,important_gene_indices[j]])
        plt.title(genes[important_gene_indices[j]]+' (%d): '%(comparisons[j])+sn(scores[j]))
        _ = plt.axis('off')

# Select features from history
def feature_extraction_via_split_depth(history):
    feature_lists = []
    for i in history:
        feature_lists.append(i[1][:np.max([20-i[-1],1])])
    return reduce(np.union1d,feature_lists)

# Remove entries from shistory that would not be there if the splitting step was performed 
# with the specified threshold
def filter_out_extraneous_steps(shistory,threshold):
    shistory_filt = [i for i in shistory if len(i) < 8 or i[6] > threshold]
    a = [sorted(i[0].keys(),key=lambda x:len(x))[0] for i in shistory_filt]
    flags = [False for i in range(len(a))]
    keep_entries = {i:None for i in a}
    # Given a list of strings, for each string s of length N, remove it if s[:-1] is not in the list
    prev_len = len(keep_entries)
    while True:
        for j,i in enumerate(keep_entries.keys()):
            if len(i) > 1 and i[:-1] not in keep_entries: keep_entries.pop(i)
        if prev_len == len(keep_entries): break
        prev_len = len(keep_entries)
    
    for j,i in enumerate(a):
        if i not in keep_entries: flags[j] = True
        
    return [i for j,i in enumerate(shistory_filt) if not flags[j]]

# Perform parameter sweeping using the above function
def get_clusters_from_history(D,shistory,threshold,disband_percentile):
    shistory = filter_out_extraneous_steps(shistory,threshold)
    
    # For a given threshold, recover the clustering from D
    N = len(D)
    labels = ['r' for i in range(N)]
    for q,i in enumerate(shistory): _,labels = update_labels(i[1],labels)
    labels = np.array(labels).astype('S%d'%(max([len(i) for i in labels])+len(str(N))))

    # Remove singletons
    disband_threshold = np.percentile(flatten_distance_matrix(D),disband_percentile)
    for i in np.unique(labels):
        inds = labels == i
        if np.sum(inds) > 1 and np.sum(flatten_distance_matrix(D,inds) < disband_threshold) == 0:
            for j in range(N):
                if inds[j]: labels[j] = labels[j]+str(j)
                    
    return labels



import numpy as np
import networkx as nx

# NMI
from sklearn.metrics import normalized_mutual_info_score
def NMI(x,y):
    return normalized_mutual_info_score(x,y)

# Max weight matching
# gets max weight matching of a biparetite graph with row_label x column_label
# (weights are given by weight_matrix)
def get_max_wt_matching(row_label,column_label, weight_matrix):
    # Create a bipartite graph where each group has |unique labels| nodes 
    G = nx.complete_bipartite_graph(len(row_label), len(column_label))
    # Weight each edge by the weight in weight matrix.. 
    for u,v in G.edges(): G[u][v]["weight"]=weight_matrix[u,v-len(row_label)]
    # Perform weight matching using Kuhn Munkres
    H=nx.max_weight_matching(G)
    max_wt=0
    for u,v in H.items(): max_wt+=G[u][v]["weight"]/float(2)
    return max_wt

def compute_clustering_accuracy(label1, label2):
    uniq1,uniq2 = np.unique(label1),np.unique(label2)
    # Create two dictionaries. Each will store the indices of each label
    entries1,entries2 = {},{}
    for label in uniq1: entries1[label] = set(np.flatnonzero((label1==label)))
    for label in uniq2: entries2[label] = set(np.flatnonzero((label2==label)))
    # Create an intersection matrix which counts the number of entries that overlap for each label combination        
    W = np.zeros((len(uniq1),len(uniq2)))
    for i,j in itertools.product(range(len(uniq1)),range(len(uniq2))):
        W[i,j]=len(entries1[uniq1[i]].intersection(entries2[uniq2[j]]))
    # find the max weight matching
    match_val = get_max_wt_matching(uniq1,uniq2,W)
    # return the error rate
    return match_val/float(len(label1))



import numpy as np
import matplotlib.pyplot as plt
from scipy import stats
from sklearn.manifold import TSNE
from sklearn.metrics.pairwise import pairwise_distances
from sklearn.decomposition import FastICA,NMF
from sklearn.preprocessing import StandardScaler

# TRANSFORMATIONS

# Remove genes with greater than some number of expression across all cells
# Also remove genes with 'MT' in its name (these are mitochondrial genes)
def filter_genes(X,genes,thresh=0):
    keep_inds_thresh = np.where(np.sum(X,0) > thresh)[0]
    keep_inds_MT = np.array([i for i in range(len(genes)) if 'MT-' not in genes[i].upper()])
    keep_inds = np.intersect1d(keep_inds_thresh,keep_inds_MT)
    print 'Kept %d features for having > %d counts across all cells'%(len(keep_inds),thresh)
    return X[:,keep_inds],genes[keep_inds]

# Downsample the number of cells
def samp_data(X,percent=50):
    n = len(X)
    inds = np.random.choice(n,int(percent/100.*n),replace=False)
    return X[inds,:]

# Map features to rankings
def feature_ranking_map(X):
    X += np.random.normal(0,0.01,np.shape(X))
    Y = np.zeros_like(X)
    for i in range(len(Y.T)):
        Y[:,i] = stats.rankdata(X[:,i], "average")
    return Y

# Normalize rows of data matrix (such that they sum to 1)
def normalize(Z): return np.nan_to_num(np.array(Z/np.matrix(np.sum(Z,1)).T))

# Take log, then normalize columns of data matrix (such that their L2 norms are 1)
def log_unit_norm_features(X):
    return X/np.linalg.norm(X,axis=0).reshape(1,-1)

# Take log, then normalize the variance of each column to 1
def log_stdvar(X):
    Z = np.log(X+1)
    return Z/np.sqrt(np.var(Z,axis=0).reshape(1,-1))

# Take log, the zero-mean and std var each column
def log_standardize(X):
    Z = np.log(X+1)
    ss = StandardScaler()
    return ss.fit_transform(Z)

# First normalize columns, then normalize rows
def normalize_all(X):
    Z = normalize(X.T)
    return normalize(Z.T)

# 10x approach: normalize across cells, multiply each cell with median, log,
#     standardize each gene
def normalizecounts_multiplymedian_log_standardize(X):
    med = np.median(np.sum(X,1))
    Z = normalize(X)*med
    return log_standardize(Z)

# For dropseq approach, do above after performing gene selection
def preprocessing_dropseq(X):
    keep_inds = dropseq_gene_selection(np.log(X+1))
    return normalizecounts_multiplymedian_log_standardize(X[:,keep_inds])

# For 10X approach, do both the above in addition to PCA
def preprocessing_10x(X,state_var_prop=False):
    Z = preprocessing_dropseq(X)
    zpca = sk_pca(X)
    if state_var_prop:
        vs = np.sum(np.square(np.linalg.norm(zpca,axis=1)))
        s = np.sum(np.square(np.linalg.norm(zpca[:,:50],axis=1)))/vs
        print 'Variance explained by first 50 pcs: %.3f'%(s)
    return zpca[:,:50]

# Dropseq gene selection 
def assign_to_bins(v,bins=20):
    last_ind = len(v)-len(v)%bins
    sorted_inds = np.argsort(v)
    bin_assignments = np.split(sorted_inds[:last_ind],bins)
    bin_assignments[-1] = np.append(bin_assignments[-1],sorted_inds[last_ind:])
    assignments = np.zeros(len(v))
    for i,b in enumerate(bin_assignments):
        assignments[b] = i
    return assignments

def dropseq_gene_selection(X,z_cutoff=1.7,bins=20):
    m = np.mean(X,0)
    assignments = assign_to_bins(m,bins=bins)
    dispersion = np.divide(np.var(X,0),m)
    # select genes to keep based on zscore within assigned bins
    keep_inds = np.zeros(np.shape(X)[1])
    for i in range(bins):
        zscores = stats.mstats.zscore(dispersion[assignments == i])
        keep_inds[assignments == i] = zscores > z_cutoff
    return keep_inds.astype('bool')

# ------------------------------------------------------------------------------
# DIMENSIONALITY REDUCTION

# sklearn's ICA
def sk_ica(X,nc=10):
    ica = FastICA(n_components=nc)
    return ica.fit_transform(X)

# sklearn's tSNE 
def sk_tsne(X,showplot=False,colors=None,title=None):
    model = TSNE(verbose=False)
    X_tSNE = model.fit_transform(X)
    if showplot:
        plot_embedding(X_tSNE[:,0],X_tSNE[:,1],colors,title+' (First two tSNE components)')
    return X_tSNE

# sklearn's tSNE with distance precomputing
def sk_tsne_precompute(D,showplot=False,colors=None,title=None):
    model = TSNE(verbose=False,metric='precomputed')
    X_tSNE = model.fit_transform(D)
    if showplot:
        plot_embedding(X_tSNE[:,0],X_tSNE[:,1],colors,title+' (First two tSNE components)')
    return X_tSNE

# Custom implementation of PCA (using SVD)
def sk_pca(X,showplot=False,colors=None,title=None):
    X_ = np.array(X)
    X_ -= np.mean(X_,0)
    U,s,V = np.linalg.svd(X_,full_matrices=0)
    pcs = np.dot(U,np.diag(s))
    if showplot:
        plt.figure()
        plt.stem(s)
        plt.xlim([0,20]) # Show first 20 singular values
        plt.title(title+' (SVD singular values)')
        plot_embedding(pcs[:,0],pcs[:,1],colors,title+' (First two PCs)')
    return pcs

# Low dimensional embedding using tSNE and PCA
def low_dimensional_embedding(X,pcs=10):
    xpca = sk_pca(X,showplot=False)
    xtsne = sk_tsne(xpca[:,:pcs],showplot=False)
    return xtsne[:,0],xtsne[:,1]

# Function for visualizing tsne or pca
def plot_embedding(x,y,colors,title):
    plt.figure()
    plt.scatter(x,y,edgecolors='none',c=colors)
    plt.colorbar()
    plt.title(title)

def compute_pcs_needed_to_explain_variance(X,percvar,verbose=False):
    U,S,V = np.linalg.svd(X)
    S = np.square(S) # Squared singular value = % variance explained by each pc
    cs = np.cumsum(S)
    thresh = cs[-1]*percvar/float(100)
    if thresh < cs[0]:
        if verbose: print 'First pc explains %.3f%% of the variance'%(100*cs[0]/cs[-1])
        return 1
    else:
        npcs = np.where(cs < thresh)[0][-1] + 1
        if verbose: print 'Number of pcs to explain %d%% of the variance: %d'%(percvar,npcs)  
        return npcs

# ------------------------------------------------------------------------------
# DISTANCE METRICS

def pairwise_L2(X): # L2
    return pairwise_distances(X)

def pairwise_L2_after_PCA(X): # L2 after a dimensionality reduction
    Xpca = sk_pca(X,showplot=False)
    return pairwise_distances(Xpca[:,:5])

def pairwise_L1(X): # L1 after normalizing each sample
    Xnorm = np.nan_to_num(normalize(X))
    return pairwise_distances(Xnorm,metric='l1')

def pairwise_correlation(X): # Pearson correlation 
    return pairwise_distances(X,metric='correlation')

def get_D_percentile(D,percentile):
    return np.percentile(D.reshape(1,-1),percentile)

# ------------------------------------------------------------------------------
# FULL PREPROCESSING OPTIONS (counts --> distance function)

# Log, standardize, PCA, remove PCS correlated with proportion of genes expressed
def log_std_PCA_filterpcs(X):
    X = log_standardize(X)
    # Compute proportion of genes expressed
    cdr = np.sum(X > 0,1)
    # PCA
    U,S,V = np.linalg.svd(X)
    pcs = np.dot(U,np.diag(S))
    # Compute correlation between each PC and CDR
    pc_cdr_corr = np.ndarray.flatten(np.abs(1-pairwise_distances(U.T,cdr.reshape(1,-1),metric='correlation')))
    # Only keep PCs which do not have high correlation with CDR
    S = S[pc_cdr_corr < 0.8]
    pcs = pcs[:,pc_cdr_corr < 0.8]
    print '%d pcs removed for having high correlation with proportion of genes expressed'%(len(U)-len(S))
    cs = np.cumsum(np.square(S))
    thresh = cs[-1]*0.8 # Arbitrarily chose 80% of variance explained
    npcs = np.where(cs < thresh)[0][-1] + 1
    X = pcs[:,:npcs]
    print '%d pcs kept to explain 80%% of the variance'%(np.shape(X)[1])
    return X

# Log, standardize variance, NMF, euclidean
def log_stdvar_NMF_L2(X):
    X = log_stdvar(X)
    k = compute_pcs_needed_to_explain_variance(X,50)
    nmf = NMF(n_components=k)
    Xrd = nmf.fit_transform(X)
    return pairwise_distances(Xrd)

# Log, standardize, PCA, L2
def log_PCA_L2(X):
    X = log_standardize(X)
    X = sk_pca(X,showplot=False)[:,:10]
    return pairwise_distances(X)

# Log, PCA, L2
def log_standardize_PCA_L2(X):
    X = np.log(X+1)
    X = sk_pca(X,showplot=False)[:,:10]
    return pairwise_distances(X)

# Normalize counts, L1
def normalizecounts_L1(X):
    X = normalize(X)
    return pairwise_L1(X)

def log_std_PCA_filterpcs_correlation(X):
    return pairwise_correlation(log_std_PCA_filterpcs(X))

def log_std_PCA_filterpcs_L2(X):
    return pairwise_L2(log_std_PCA_filterpcs(X))

# Log, L1
def log_L1(X):
    return pairwise_L1(np.log(X+1))

# Log, normalize, L1
def log_norm_L1(X):
    X = normalize(np.log(X+1))
    return pairwise_L1(X)

# Log, standardize, L1
def log_standardize_L1(X):
    X = log_standardize(X)
    return pairwise_L1(X)

# Log, unit var, L1
def log_stdvar_L1(X):
    X = log_stdvar(X)
    return pairwise_L1(X)

# Log, correlation
def log_correlation(X):
    return pairwise_correlation(np.log(X+1))

# Log, normalize, correlation
def log_norm_correlation(X):
    X = normalize(np.log(X+1))
    return pairwise_correlation(X)

# Log, standardize, correlation
def log_standardize_correlation(X):
    X = log_standardize(X)
    return pairwise_correlation(X)

# Log, unit var, correlation
def log_stdvar_correlation(X):
    X = log_stdvar(X)
    return pairwise_correlation(X)



import numpy as np
import networkx as nx
import community,time,itertools,pickle
from scipy.stats import ttest_ind
from scipy.spatial.distance import squareform
from utils import *
from preprocessing import *
from feature_selection import *

# Clean-up step: For each cluster that's too small, find the cluster that it's closest to and 
# merge those clusters together

# Approach: Pairwise distances between clusters. 
# Distance = mininum p-value between the two?
# If two clusters are similar, they should have a small distance. There will be few genes that distinguish the two
# and a small -log10 p-value will be generated (distributions of all genes are similar). 

# 1. Look at the pool of small clusters. 
# 2. Choose the one that has the lowest distance to any cluster. 
# 3. Merge those clusters. Recompute the matrix
# 4. Repeat

def find_min(D):
    n = len(D)
    D = np.multiply(D,np.tril(np.ones([n,n]),-1))+np.max(D)*np.triu(np.ones([n,n]))
    # in case of ties, just pick the first one that's not along a diagonal
    min_inds = np.where(D == np.min(D)) 
    for i in range(len(min_inds[0])):
        if min_inds[0][i] != min_inds[1][i]:
            return np.sort([min_inds[0][i],min_inds[1][i]])

# Compute the distance between clusters using minimum p-value for any gene t-test
def cdist_log_Welchs(X,D,i,j,z,cache):
    X1 = X[z == i]
    X2 = X[z == j]
    # Singleton case: return largest int (let singleton remain as outlier)
    if len(X1) == 1 or len(X2) == 1: return np.nan_to_num(np.inf)
    X = np.vstack((X1,X2))
    y = np.zeros(len(X))
    y[len(X1):] = 1
    _,_,score = log_select_genes_using_Welchs(X,y,return_score=True)
    return cache,score

# Compute the distance between clusters using sigclust
def cdist_log_sigclust(X,D,i,j,z,cache):
    X1 = np.log(1+X[z==i])
    X2 = np.log(1+X[z==j])
    X = np.vstack((X1,X2))
    y = np.zeros(len(X))
    y[len(X1):] = 1
    score = sigclust(X,y)
    return cache,score

# Given bins and some samples, compute the distribution over those bins
def dist_over_bins(x,bins):
    num_bins = len(bins)-1
    dist = np.bincount(np.digitize(x,bins),minlength=num_bins+1)[1:]
    return dist.astype(float)/np.sum(dist)

# Compute the distance between clusters using distribution of correlation bins after merge
def cdist_corr_distributions(X,D,i,j,z,cache,show_plot=False):
    
    num_bins = 20
    if 'bins' not in cache:
        print 'computing bins'
        d = flatten_distance_matrix(D)
        bins = np.percentile(d,np.linspace(0,100,num_bins+1))
        bins[-1] = np.inf
        cache['bins'] = bins
        
    dM = flatten_distance_matrix(D,np.logical_or(z==i,z==j))
    distM = dist_over_bins(dM,cache['bins'])
    distance = np.sum(distM[-num_bins/2:])
    
    if show_plot:
        plt.stem(distM)
        plt.xlim((-1,num_bins))
        plt.title('nL:%d, nR:%d, nM:%d, dist: ' \
                  %(np.sum(z==i),np.sum(z==j),np.sum(z!='_'))+sn(distance))
    
    return cache,distance

# Compute the distance between clusters using proportion of clustering in the bottom percentile
# Does the same thing as above function, only a bit faster
def cdist_corr_percentile(X,D,i,j,z,cache):
    if 'cutoff' not in cache:
        d = flatten_distance_matrix(D)
        cache['cutoff'] = np.percentile(D,50)
    dM = flatten_distance_matrix(D,np.logical_or(z==i,z==j))
    distance = np.sum(dM[dM > cache['cutoff']])
    return cache,distance

# visualize distributions of distances in each cluster
def visualize_within_cluster_distance_distributions(D,y,num_bins=20,num_plots_per_row=3,show_D_dist=False):
    d = flatten_distance_matrix(D)
    bins = np.percentile(d,np.linspace(0,100,num_bins+1))
    if show_D_dist:
        plt.figure(figsize=(16,4))
        h = plt.hist(d,bins=num_bins)
        for i in range(num_bins+1):
            plt.plot([bins[i],bins[i]],[0,np.max(h[0])*1.1],'r')
        plt.ylim([0,np.max(h[0])*1.1])
        plt.title('Histogram of distances for all pairs of points')
    bins[-1] = np.inf
    j = 0
    for i in np.unique(y):
        if j%num_plots_per_row == 0: plt.figure(figsize=(16,2))
        if np.sum(y==i) == 1: continue
        plt.subplot(1,num_plots_per_row,j%num_plots_per_row+1)
        dM = flatten_distance_matrix(D,y==i)
        distM = dist_over_bins(dM,bins)
        plt.stem(distM)
        plt.xlim((-1,num_bins))
        plt.title('C:%d, N=%d, score='%(i,np.sum(y==i))+sn(np.sum(distM[-num_bins/2:])))
        j += 1

# Map labels to integers between 0 and #labels-1
def remap_labels(L):
    c = np.unique(L)
    mapper = {c[i]:i for i in range(len(c))}
    return np.array([mapper[i] for i in L])

# Merge function for handling singletons: assign to same cluster as nearest neighbor
def assign_singletons(X,y,preprocessing=log_correlation,verbose=False,
                      allow_singleton_pairs=True,threshold_percentile=None):
    z = str_labels_to_ints(y)
    z_counts = np.bincount(z)
    Ns = np.sum(z_counts == 1)
    if verbose: print '%d of %d samples are singletons'%(Ns,len(y))
    if Ns == 0: return z
    if preprocessing != 'precomputed': 
        D = preprocessing(X)
        D_backup = np.copy(D)
    else: D = X

    # Save a list of outliers (points that are too far from all other points according to some threshold)
    # by putting them in cluster "-1"
    if threshold_percentile is None: threshold = np.max(D)
    else: threshold = np.percentile(flatten_distance_matrix(D),threshold_percentile)
    if verbose: print 'Outlier threshold is '+sn(threshold)

    # Truncate distance matrix so that is a Ns-by-N matrix of distances
    sing_ind = np.zeros(len(y))
    for i in range(len(z_counts)):
        if z_counts[i] == 1: sing_ind[z == i] = 1
    sing_ind = sing_ind.astype(bool)
    sing_to_orig_map = np.array(range(len(y)))[sing_ind]
    if allow_singleton_pairs:
        for i in range(len(D)): D[i,i] = np.inf # Ignore diagonal
    D = D[sing_ind,:]

    # Account for case where we don't want a singleton being assigned to another singleton
    if not allow_singleton_pairs:
        nonsing_to_orig_map = np.array(range(len(y)))[np.invert(sing_ind)]
        D = D[:,np.invert(sing_ind)]

    # Grab nearest neighbor for each singleton
    nneighbor = np.argmin(D,1)

    def is_singleton_str(x,sid):
        if sid[x]: return ' (singleton)'
        return ''

    if allow_singleton_pairs:
        nneighbor_dists = np.min(D,1)
        for(d,i) in sorted([(dn,j) for j,dn in enumerate(nneighbor_dists)]):
            if verbose: print '%d\'s nearest neighbor: %d%s in cluster %d (D = %.3f)' \
                              %(sing_to_orig_map[i],nneighbor[i],
                                is_singleton_str(nneighbor[i],sing_ind),
                                z[nneighbor[i]],D[i,nneighbor[i]])
            if D[i,nneighbor[i]] < threshold:
                z[z == z[sing_to_orig_map[i]]] = z[nneighbor[i]]
            else: z[z == z[sing_to_orig_map[i]]] = -1

    else:
        for i in range(len(nneighbor)):
            if verbose: print '%d\'s nearest neighbor: %d%s in cluster %d (D = %.3f)' \
                              %(sing_to_orig_map[i],nneighbor[i],
                                is_singleton_str(nneighbor[i],sing_ind),
                                z[nneighbor[i]],D[i,nneighbor[i]])
            if D[i,nneighbor[i]] < threshold:
                z[sing_to_orig_map[i]] = z[nonsing_to_orig_map[nneighbor[i]]]
            else: z[z == z[sing_to_orig_map[i]]] = -1
    if verbose: print 'Total number of outliers: %d'%(np.sum(z==-1))
    z_out = remap_labels(z)
    if np.sum(z==-1) > 0: z_out -= 1
    if preprocessing != 'precomputed': return D_backup,z_out
    return z_out
    
def dendromerge(X,y,score_threshold=None,preprocessing=log_correlation,clust_dist=cdist_log_Welchs,
                verbose=False,split_evaluator=log_select_genes_using_Welchs,save_prefix=None,
                outlier_threshold_percentile=100,perform_community_detection=False,return_Dc_history=False):
    '''
    Clean-up step following the splitting step. Merges clusters basd on some criteria.

    Inputs:
    X: input matrix of molecule counts. (D,X) tuple if distance matrix precomputed
    y: labels of clusters 
    score_threshold: threshold of score to determine if we should stop merging. i.e. if all clusters
       are at least this far away from each other, stop merging clusters
    preprocessing: function to generate distance matrix from X
    clust_dist: function to determine the distance between clusters
    verbose: whether or not to print the details of the merging phase
    split_evaluator: function to determine how well separated two clusters are (used for history)
    save_prefix: if not None, then the pairwise cluster distance matrix is saved
    outlier_threshold_percentile: if a singleton's distance to its closest point is in the top
       1-outlier_threshold_percentile of overall distances, then mark the singleton as an outlier
    perform_community_detection: boolean to determine whether or not to merge using community
       detection on the pairwise cluster distance matrix instead

    Outputs:
    ym: labels after merging
    history: first entry is the set of original labels, second entry is the set of labels after
       assigning singletons. Every entry after that tracks some information about each merge in the
       same way as the splitting step.
    '''

    start_time = time.time()

    if score_threshold is None and perform_community_detection is False:
        print 'Please enter a score threshold for the score_threshold keyword'
        return

    if preprocessing != 'precomputed':
        D,z = assign_singletons(X,y,preprocessing=preprocessing,verbose=verbose,
                                threshold_percentile=outlier_threshold_percentile)
        if verbose: print 'Preprocessing took %.3f s'%(time.time()-start_time)
    elif type(X) != tuple:
        print 'preprocessing = precomputed. Feed in (D,X) tuple as first argument.'
        return
    elif not np.all(np.isclose(X[0],X[0].T)):
        print 'Need a valid distance matrix.'
        return
    else: 
        D,X = X[0],X[1]
        z = assign_singletons(np.copy(D),y,preprocessing=preprocessing,verbose=verbose,
                              threshold_percentile=outlier_threshold_percentile)
    if verbose: print 'Singletons assigned (%.3f s)'%(time.time()-start_time)

    # Handle outliers (0's from previous step)
    outliers = np.where(z==-1)[0]
    non_outliers = np.where(z!=-1)[0]
    D = cut_matrix_along_both_axes(D,z!=-1)
    X = X[z!=-1,:]
    z = remap_labels(z[z!=-1])

    # Save useful things to not have to recompute
    cache = {}

    N_samples = len(outliers)+len(non_outliers)
    Nc = len(np.unique(z))
    Dc = np.zeros((Nc,Nc))
    N,M = np.shape(X)

    # Map outlier-only labels to original labels
    def readjust_labels(L,padval):
        L_adjust = np.array([padval for i in range(N_samples)])
        for i,non_outlier in enumerate(non_outliers):
            L_adjust[non_outlier] = L[i]
        return L_adjust

    history = [str_labels_to_ints(y),readjust_labels(z,-1)]

    # Compute distances for all clusters
    for i,j in itertools.combinations(range(Nc),2):
        cache,Dc[i,j] = clust_dist(X,D,i,j,z,cache)
        Dc[j,i] = Dc[i,j]

    if save_prefix is not None: pickle.dump(Dc,file(save_prefix+'_Dc.pickle','w'))
    if verbose: print 'Dc generated (%.3f s)'%(time.time()-start_time)

    # Edge case
    if len(Dc) < 2: return readjust_labels(z,-1),history

    # Merge using community detection
    if perform_community_detection:

            
        # Convert distance matrix to edge tuples for networkx
        def D_to_squareform_with_thresholding(D,thresh=None):
            indices = list(itertools.combinations(range(0,len(D)),2))
            if thresh is None: thresh = np.percentile(flatten_distance_matrix(D),100./np.sqrt(len(D)))
            else: thresh = np.percentile(flatten_distance_matrix(D),thresh)
            D_squareform = []
            for j,i in enumerate(squareform(D)):
                if i < thresh: D_squareform.append(indices[j])
            return D_squareform
        
        edge_weights = D_to_squareform_with_thresholding(Dc)
        G = nx.Graph()
        G.add_nodes_from(np.unique(z))
        G.add_edges_from(edge_weights)
        if verbose: print 'Graph constructed with %d nodes and %d edges.. (%.3f s)' \
                    %(len(G.nodes()),len(G.edges()),time.time()-start_time)
        partition = community.best_partition(G)
        if verbose: print 'Merging clusters took %.3f s'%(time.time()-start_time)
        return readjust_labels(np.array([partition[i] for i in z]),-1),history

    # Original merge algorithm: iteratively merge the two clusters with the smallest distance
    merge_depth = 1
    while len(Dc) > 1:
        merge_inds = find_min(Dc)
        if Dc[merge_inds[0],merge_inds[1]] > score_threshold: break

        if verbose:
            print 'Before the merge: %d clusters'%(Nc)
            print 'Merging labels %d (N = %d) and %d (N = %d) with distance ' \
                  %(merge_inds[0],np.sum(z==merge_inds[0]),merge_inds[1],
                    np.sum(z==merge_inds[1])) + sn(Dc[merge_inds[0],merge_inds[1]])
        
        # Get info for history
        y_node = np.array(['_' for i in range(N)])
        y_node[z == merge_inds[0]] = 'L'
        y_node[z == merge_inds[1]] = 'R'
        genes,rankings,score,comparisons = split_evaluator(X[y_node!='_',:],y_node[y_node!='_'],
                                                           return_score=True,return_comparisons=True)
        history.append(({'m': np.sum(y_node != '_'),
                         'mL': np.sum(y_node == 'L'),
                         'mR': np.sum(y_node == 'R')},
                         readjust_labels(y_node,'_'),Dc[merge_inds[0],merge_inds[1]],merge_depth,
                         genes, rankings, score, comparisons))
            
        # Delete the rows/columns of the distance matrix corresponding to the two merged clusters
        Dc = np.delete(np.delete(Dc,merge_inds,1),merge_inds,0)

        # Merge the clusters
        z[z == merge_inds[0]] = Nc
        z[z == merge_inds[1]] = Nc
        z = remap_labels(z)
        Nc -= 1

        # Create new distance matrix
        dc = np.zeros(Nc-1)
        for i in range(Nc-1):
            cache,dc[i] = clust_dist(X,D,i,Nc-1,z,cache)
        dc = np.matrix(dc)
        Dc = np.vstack((np.hstack((Dc,dc.T)),np.hstack((dc,np.matrix(0)))))
        
        merge_depth += 1

    # Readjust z to include the original outliers
    if verbose: print 'Merging clusters took %.3f s'%(time.time()-start_time)
    return readjust_labels(z,-1),history



from distutils.core import setup
setup(
  name = 'dendrosplit',
  packages = ['dendrosplit'], # this must be the same as the name above
  version = '0.1',
  description = 'A Python 2.7 package for performing interpretable clustering and feature selection for single-cell RNA-Seq datasets.',
  author = 'Jesse M. Zhang, Jue Fan, H. Christina Fan, David Rosenfeld, David Tse',
  author_email = 'jessez@stanford.edu',
  url = 'https://github.com/jessemzhang/dendrosplit', # use the URL to the github repo
  download_url = 'https://github.com/jessemzhang/dendrosplit/archive/0.1.tar.gz',
  keywords = ['single-cell', 'rna-seq', 'clustering'], # arbitrary keywords
  classifiers = [],
)
